// -*- Mode: C++ -*-

// ----------------------------------------------------------------------------
//  map_file.cpp
// ----------------------------------------------------------------------------

// ----------------------------------------------------------------------------
//  CN[h
// ----------------------------------------------------------------------------

#include "map_file.h"
#include "util.h"


// ----------------------------------------------------------------------------
//  }bvgt@C
// ----------------------------------------------------------------------------

int map_file::create(HANDLE fh, DWORD protect, DWORD size, const char *name)
{
  close();
  m_fmo = CreateFileMapping(fh,
                            NULL,
                            protect,
                            0, size,
                            name);
  return (NULL == m_fmo) ? 1 : 0;
}


int map_file::open(DWORD acc, const char *name)
{
  close();
  m_fmo = OpenFileMapping(acc, FALSE, name);
  return (NULL == m_fmo) ? 1 : 0;
}


void map_file::close(void)
{
  if (NULL != m_fmo)
  {
    unmap();
    CloseHandle(m_fmo);
    m_fmo = NULL;
  }
}


void *map_file::map(DWORD acc)
{
  unmap();
  m_addr = MapViewOfFile(m_fmo, acc, 0, 0, 0);
  return m_addr;
}


void map_file::unmap(void)
{
  if (NULL != m_addr)
  {
    UnmapViewOfFile(m_addr);
    m_addr = NULL;
  }
}


// ----------------------------------------------------------------------------
//  (Ot) t@C}bsOIuWFNg
// ----------------------------------------------------------------------------

int file_mapping_object::create(const char *name, size_t data_size)
{
  close();

  if (0 == data_size)
    return 1;

  if (m_fmo.create(INVALID_HANDLE_VALUE, PAGE_READWRITE,
                   data_size + sizeof(size_t), name))
    return 1;

  void  *ptr = m_fmo.map(FILE_MAP_WRITE);
  if (NULL == ptr)
  {
    m_fmo.close();
    return 1;
  }

  // TCY
  m_size = data_size;
  memcpy(ptr, &m_size, sizeof(size_t));
  m_fmo.unmap();

  return 0;
}


int file_mapping_object::open(const char *name)
{
  close();

  if (m_fmo.open(FILE_MAP_READ, name))
    return 1;

  void  *ptr = m_fmo.map(FILE_MAP_READ);
  if (NULL == ptr)
  {
    m_fmo.close();
    return 1;
  }

  // TCYǂݍ
  memcpy(&m_size, ptr, sizeof(size_t));
  m_fmo.unmap();

  return 0;
}


void file_mapping_object::close(void)
{
  m_size = 0;
  m_fmo.close();
}


int file_mapping_object::write(const void *data, size_t write_size)
{
  size_t  *ptr = (size_t *)m_fmo.map(FILE_MAP_WRITE);
  if (NULL == ptr)
    return 1;

  memcpy(ptr + 1, data, MIN(m_size, write_size));
  m_fmo.unmap();

  return 0;
}


int file_mapping_object::read(void *buf, size_t buf_size)
{
  const size_t  *ptr = (size_t *)m_fmo.map(FILE_MAP_READ);
  if (NULL == ptr)
    return 1;

  memcpy(buf, ptr + 1, MIN(m_size, buf_size));

  return 0;
}

