// -*- Mode: C++ -*-

// ----------------------------------------------------------------------------
//  mspack_if.cpp
// ----------------------------------------------------------------------------

// ----------------------------------------------------------------------------
//  CN[h
// ----------------------------------------------------------------------------

#include "mspack_if.h"
#include "util.h"
#include "map_file.h"

#include <stdio.h>


// ----------------------------------------------------------------------------
//  vg^Cv錾
// ----------------------------------------------------------------------------

extern "C" struct mspack_file *msp_open(struct mspack_system *, char *, int);
extern "C" void msp_close(struct mspack_file *);
extern "C" int msp_read(struct mspack_file *, void *, int);
extern "C" int msp_write(struct mspack_file *, void *, int);
extern "C" int msp_seek(struct mspack_file *, off_t, int);
extern "C" off_t msp_tell(struct mspack_file *);
extern "C" void msp_message(struct mspack_file *, char *, ...);
extern "C" void *msp_alloc(struct mspack_system *, size_t);
extern "C" void msp_free(void *);
extern "C" void msp_copy(void *, void *, size_t);


// ----------------------------------------------------------------------------
//  axcab p mspack_system 擾
// ----------------------------------------------------------------------------

struct mspack_system *get_mspack_system(void)
{
  static struct mspack_system  msp_sys =
  {
    msp_open,
    msp_close,
    msp_read,
    msp_write,
    msp_seek,
    msp_tell,
    msp_message,
    msp_alloc,
    msp_free,
    msp_copy,
    NULL,
  };

  return &msp_sys;
}


// ----------------------------------------------------------------------------
//  t@Co̓C^tFCX
// ----------------------------------------------------------------------------

class axcab_mspack_file
{
public:
  axcab_mspack_file(void) { }
  virtual ~axcab_mspack_file() { }

  virtual int open(const char *filename, int mode) = 0;
  virtual void close(void) = 0;
  virtual int read(void *buffer, int bytes) = 0;
  virtual int write(const void *buffer, int bytes) = 0;
  virtual int seek(off_t offset, int mode) = 0;
  virtual off_t tell(void) = 0;
};


// ----------------------------------------------------------------------------
//  t@C
// ----------------------------------------------------------------------------

#if 0  // Wo͔

class disk_file : public axcab_mspack_file
{
  FILE  *m_fp;

  static int get_mode_num(int mode)
  {
    switch (mode)
    {
    case MSPACK_SYS_OPEN_READ:   return 0;
    case MSPACK_SYS_OPEN_WRITE:  return 1;
    case MSPACK_SYS_OPEN_UPDATE: return 2;
    case MSPACK_SYS_OPEN_APPEND: return 3;
    default: return -1;
    }
  }

public:
  disk_file(void)
       : m_fp(NULL) { }
  ~disk_file() { close(); }

  int open(const char *filename, int mode)
  {
    close();
    if (0 > (mode = get_mode_num(mode)))
      return 1;

    static const char  *MODE_STR[] = { "rb", "wb", "r+b", "ab" };
    m_fp = fopen(filename, MODE_STR[mode]);

    return (NULL == m_fp);
  }

  void close(void)
  {
    if (NULL != m_fp)
    {
      fclose(m_fp);
      m_fp = NULL;
    }
  }

  int read(void *buffer, int bytes)
  {
    return fread(buffer, 1, bytes, m_fp);
  }

  int write(const void *buffer, int bytes)
  {
    return fwrite(buffer, 1, bytes, m_fp);
  }

  int seek(off_t offset, int mode)
  {
    switch (mode)
    {
    case MSPACK_SYS_SEEK_START: mode = SEEK_SET; break;
    case MSPACK_SYS_SEEK_CUR:   mode = SEEK_CUR; break;
    case MSPACK_SYS_SEEK_END:   mode = SEEK_END; break;
    default: return -1;
    }
    return fseek(m_fp, offset, mode);
  }

  off_t tell(void) { return ftell(m_fp); }
};

#else  // Win32 API 

class disk_file : public axcab_mspack_file
{
  HANDLE  m_fh;
  DWORD  m_io_bytes;

public:
  disk_file(void)
       : m_fh(INVALID_HANDLE_VALUE) { }
  ~disk_file() { close(); }

  int open(const char *filename, int mode)
  {
    close();

    int access_mode, share_mode, create_mode;
    switch (mode)
    {
    case MSPACK_SYS_OPEN_READ:
      access_mode = GENERIC_READ;
      share_mode = FILE_SHARE_READ;
      create_mode = OPEN_EXISTING;
      break;
    default:
      return 1;
    }

    m_fh = CreateFile(filename,
                      access_mode,
                      share_mode,
                      NULL,
                      create_mode,
                      FILE_ATTRIBUTE_NORMAL,
                      NULL);

    if (INVALID_HANDLE_VALUE == m_fh)
      return 1;

    return 0;
  }

  void close(void)
  {
    if (INVALID_HANDLE_VALUE != m_fh)
    {
      CloseHandle(m_fh);
      m_fh = INVALID_HANDLE_VALUE;
    }
  }

  int read(void *buffer, int bytes)
  {
    ReadFile(m_fh, buffer, bytes, &m_io_bytes, NULL);
    return m_io_bytes;
  }

  int write(const void *buffer, int bytes)
  {
    WriteFile(m_fh, buffer, bytes, &m_io_bytes, NULL);
    return m_io_bytes;
  }

  int seek(off_t offset, int mode)
  {
    switch (mode)
    {
    case MSPACK_SYS_SEEK_START: mode = FILE_BEGIN; break;
    case MSPACK_SYS_SEEK_CUR:   mode = FILE_CURRENT; break;
    case MSPACK_SYS_SEEK_END:   mode = FILE_END; break;
    default: return -1;
    }
    return (0xFFFFFFFF == SetFilePointer(m_fh, offset, NULL, mode)) ? -1 : 0;
  }

  off_t tell(void) { return SetFilePointer(m_fh, 0, NULL, FILE_CURRENT); }
};

#endif


// ----------------------------------------------------------------------------
//  obt@
// ----------------------------------------------------------------------------

class memory_buffer : public axcab_mspack_file
{
  uint8  *m_buf_p, *m_buf_begin;
  size_t  m_buf_size, m_buf_remain;
  spi_callback  *m_spicb;

public:
  memory_buffer(void)
       : m_buf_p(NULL),
         m_buf_begin(NULL),
         m_buf_size(0),
         m_buf_remain(0) { }
  ~memory_buffer() { close(); }

  int open(const char *filename, int mode)
  {
    close();

    file_mapping_object  fmo;
    if (fmo.open(filename))
      return 1;

    extract_info  ex_info;
    if (fmo.read(&ex_info, sizeof(ex_info)))
      return 1;

    m_buf_p = m_buf_begin = ex_info.m_buf;
    m_buf_remain = m_buf_size = ex_info.m_buf_size;
    m_spicb = ex_info.m_spicb;

    return 0;
  }

  void close(void)
  {
    m_buf_p = m_buf_begin = NULL;
    m_buf_remain = m_buf_size = 0;
  }

  int read(void *buffer, int bytes)
  {
    return -1;
  }

  int write(const void *buffer, int bytes)
  {
    size_t  copy_size = MIN(m_buf_remain, (uint)bytes);

    memcpy(m_buf_p, buffer, copy_size);
    m_buf_p += copy_size;
    m_buf_remain -= copy_size;

    return copy_size;
  }

  int seek(off_t offset, int mode)
  {
    off_t  new_pos;

    switch (mode)
    {
    case MSPACK_SYS_SEEK_START:
      new_pos = offset;
      break;

    case MSPACK_SYS_SEEK_CUR:
      new_pos = (m_buf_p - m_buf_begin) + offset;
      break;

    case MSPACK_SYS_SEEK_END:
      new_pos = m_buf_size + offset;
      break;

    default:
      return -1;
    }

    if (new_pos < 0 || m_buf_size < (uint)new_pos)
      return -1;

    m_buf_p = m_buf_begin + new_pos;
    m_buf_remain = m_buf_size - new_pos;

    return 0;
  }

  off_t tell(void) { return m_buf_p - m_buf_begin; }
};


// ----------------------------------------------------------------------------
//  name ɑΉt@C}bsOIuWFNg݂邩
// ----------------------------------------------------------------------------

int is_exist_fmo(const char *name)
{
  file_mapping_object  fmo;
  return (0 == fmo.open(name));
}


// ----------------------------------------------------------------------------
//  mspack_system C^tFCX
// ----------------------------------------------------------------------------

extern "C"
struct mspack_file *msp_open(struct mspack_system *,
                                   char *filename,
                                   int mode)
{
  axcab_mspack_file  *msp_file = NULL;
  switch (mode)
  {
  case MSPACK_SYS_OPEN_WRITE:  // Cabinet ̃t@C̒o
    if (is_exist_fmo(filename))
    {
      msp_file = new memory_buffer();
      break;
    }
    // fall thru

  case MSPACK_SYS_OPEN_READ:
  case MSPACK_SYS_OPEN_UPDATE:
  case MSPACK_SYS_OPEN_APPEND:
    msp_file = new disk_file;
    break;
  }

  if (NULL != msp_file)
  {
    if (msp_file->open(filename, mode))
    {
      delete msp_file;
      msp_file = NULL;
    }
  }

  return (struct mspack_file *)msp_file;
}


extern "C"
void msp_close(struct mspack_file *file)
{
  ((axcab_mspack_file *)file)->close();
  delete ((axcab_mspack_file *)file);
}


extern "C"
int msp_read(struct mspack_file *file, void *buffer, int bytes)
{
  return ((axcab_mspack_file *)file)->read(buffer, bytes);
}


extern "C"
int msp_write(struct mspack_file *file, void *buffer, int bytes)
{
  return ((axcab_mspack_file *)file)->write(buffer, bytes);
}


extern "C"
int msp_seek(struct mspack_file *file, off_t offset, int mode)
{
  return ((axcab_mspack_file *)file)->seek(offset, mode);
}


extern "C"
off_t msp_tell(struct mspack_file *file)
{
  return ((axcab_mspack_file *)file)->tell();
}


extern "C"
void msp_message(struct mspack_file *, char *, ...) { }


extern "C"
void *msp_alloc(struct mspack_system *,
            size_t bytes)
{
  return LocalAlloc(LMEM_FIXED | LMEM_ZEROINIT, bytes);
}


extern "C"
void msp_free(void *ptr)
{
  if (NULL != ptr)
    LocalFree(ptr);
}


extern "C"
void msp_copy(void *src, void *dest, size_t bytes)
{
  memcpy(dest, src, bytes);
}

