// -*- Mode: C++ -*-

// ----------------------------------------------------------------------------
//  axcab.spi - Cabinet extract plug-in
//  (C) 2004 mij4x
//
//  This library is free software; you can redistribute it and/or
//  modify it under the terms of the GNU Lesser General Public
//  License as published by the Free Software Foundation; either
//  version 2.1 of the License, or (at your option) any later version.
//
//  This library is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
//  Lesser General Public License for more details.

//  You should have received a copy of the GNU Lesser General Public
//  License along with this library; if not, write to the Free Software
//  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
// ----------------------------------------------------------------------------

// ----------------------------------------------------------------------------
//  spi_main.cpp
// ----------------------------------------------------------------------------

// ----------------------------------------------------------------------------
//  CN[h
// ----------------------------------------------------------------------------

#include "cab_arc.h"
#include "spi_misc.h"
#include "util.h"
#include "seq_num.h"


// ----------------------------------------------------------------------------
//  萔
// ----------------------------------------------------------------------------

#define PLUGIN_VERSION  "0.1"

const char *PLUGIN_INFO[] =
{
  "00AM",
  "Cabinet extract plug-in ver." PLUGIN_VERSION " (C) 2004 mij4x",
  "*.cab",
  "Cabinet file",
};


// ----------------------------------------------------------------------------
//  O[oϐ
// ----------------------------------------------------------------------------

HINSTANCE  g_inst = NULL;


// ----------------------------------------------------------------------------
//  Gg|Cg
// ----------------------------------------------------------------------------

BOOL WINAPI DllMain(HINSTANCE inst, DWORD reason, LPVOID)
{
  switch (reason)
  {
  case DLL_PROCESS_ATTACH:
    g_inst = inst;
    seq_num_generator::get_instance();  // 炩߃CX^X𐶐Ă
    break;

  case DLL_PROCESS_DETACH:
    break;
  }
  return TRUE;
}


// ----------------------------------------------------------------------------
//  Plug-in Ɋւ𓾂
// ----------------------------------------------------------------------------

int SPI_EXPORT GetPluginInfo(int info_no, LPSTR buf, int buf_len)
{
  int  copy_len = 0;

  if (0 <= info_no && info_no < countof(PLUGIN_INFO))
  {
    const char  *src = PLUGIN_INFO[info_no];
    char  *dst = buf;

    while (copy_len < buf_len)
    {
      copy_len++;
      if ('\0' == (*dst++ = *src++))
        break;
    }
  }

  return copy_len;
}


// ----------------------------------------------------------------------------
//  WJ\ (ΉĂ) t@C`ׂ
// ----------------------------------------------------------------------------

int SPI_EXPORT IsSupported(LPSTR filename, DWORD dw)
{
  //bool  file_handle;
  {
    const uint8  *p;
    const size_t  BUF_SIZE = 16;
    uint8  buf[BUF_SIZE];

    if (dw & 0xffff0000)
    {
      p = (const uint8 *)dw;
      //file_handle = false;
    }
    else
    {
      DWORD  read_bytes;
      ReadFile((HANDLE)dw, buf, BUF_SIZE, &read_bytes, NULL);
      p = buf;
      //file_handle = true;
    }

    if (cab_archive::is_cabinet(p))
      return 1;
  }

  return 0;

#if 0
  {
    HANDLE  fh;
    if (file_handle)
      fh = (HANDLE)dw;
    else
    {
      if (NULL == filename || '\0' == filename[0])
        return 0;

      fh = CreateFile(filename, GENERIC_READ, FILE_SHARE_READ, NULL,
                      OPEN_EXISTING, FILE_ATTRIBUTE_NORMAL, NULL);

      if (INVALID_HANDLE_VALUE == fh)
        return 0;
    }

    int  ret = cab_archive::is_cabinet_sfx(fh);

    if (!file_handle)
      CloseHandle(fh);

    return ret;
  }
#endif
}


// ----------------------------------------------------------------------------
//  A[JCûׂẴt@C̏擾
// ----------------------------------------------------------------------------

int SPI_EXPORT GetArchiveInfo(LPSTR buf, long len, unsigned int flag, HLOCAL *info)
{
  if (!is_file_in(flag))
    return SPI_ERROR_NOT_IMPLEMENTED;

  uint  count = 0;
  {
    cab_archive  arc_man;
    if (arc_man.open_archive(buf))
      return SPI_ERROR_UNKNOWN_FORMAT;

    while (0 == arc_man.get_next_fileinfo(NULL))
      count++;

    arc_man.close_archive();
  }

  fileInfo  *fi = (fileInfo *)LocalAlloc(LMEM_FIXED, sizeof(fileInfo) * (count + 1));
  if (NULL == fi)
    return SPI_ERROR_MEMORY_ALLOCATION;

  {
    cab_archive  arc_man;
    if (arc_man.open_archive(buf))
    {
      LocalFree(fi);
      return SPI_ERROR_UNKNOWN_FORMAT;
    }

    uint  i;
    for (i = 0; i < count; i++)
      if (0 != arc_man.get_next_fileinfo(&fi[i]))
        break;
    fi[i].method[0] = '\0';
//    fi[count].method[0] = '\0';

    *info = (HLOCAL)fi;

    arc_man.close_archive();
  }

  return SPI_OK;
}


// ----------------------------------------------------------------------------
//  A[JCu̎w肵t@C̏擾
// ----------------------------------------------------------------------------

int SPI_EXPORT GetFileInfo(LPSTR buf, long len, LPSTR filename, unsigned int flag, fileInfo *info)
{
  if (!is_file_in(flag))
    return SPI_ERROR_NOT_IMPLEMENTED;

  cab_archive  arc_man;
  if (arc_man.open_archive(buf))
    return SPI_ERROR_UNKNOWN_FORMAT;

  char  target_filename[200];
  char  target_path[200];
  split_filepath(filename, target_path, 200, target_filename, 200);

  int (__cdecl *str_compare)(const char *, const char *);
  str_compare = (is_ignore_case(flag)) ? std::stricmp : std::strcmp;

  fileInfo  fi;
  while (0 == arc_man.get_next_fileinfo(&fi))
  {
    if (0 == str_compare(target_path, fi.path) &&
        0 == str_compare(target_filename, fi.filename))
    {
      if (NULL != info)
        *info = fi;
      return SPI_OK;
    }
  }

  return SPI_ERROR_UNKNOWN_FORMAT;
}


// ----------------------------------------------------------------------------
//  A[JCũt@C擾
// ----------------------------------------------------------------------------

int SPI_EXPORT GetFile(LPSTR src, long len, LPSTR dest, unsigned int flag, FARPROC prog_callback, long data)
{
  if (!is_file_in(flag))// || is_file_out(flag))
    return SPI_ERROR_NOT_IMPLEMENTED;

  spi_callback  spicb((SPI_CALLBACK)prog_callback, data);

  cab_archive  arc_man;
  if (arc_man.open_archive(src))
    return SPI_ERROR_UNKNOWN_FORMAT;

  for (uint i = len; i-- > 0; )
    arc_man.get_next_fileinfo(NULL);

  fileInfo  fi;
  arc_man.get_next_fileinfo(&fi);

  int  ret;
  uint8  *img = NULL;
  if (fi.filesize)
  {
    img = (uint8 *)LocalAlloc(LMEM_FIXED, fi.filesize);
    if (NULL == img)
      return SPI_ERROR_MEMORY_ALLOCATION;

    ret = arc_man.extract(img, fi.filesize, &spicb);
    if (ret)
    {
      ret = SPI_ERROR_MEMORY_ALLOCATION;
      LocalFree(img);
      img = NULL;
    }
  }

  if (!is_file_out(flag))  // o
  {
    *(HLOCAL *)dest = (HLOCAL)img;
  }
  else if (NULL != img)  // t@Co
  {
    const size_t  BUF_SIZE = 1024;
    char  fn_buf[BUF_SIZE];

    if (0 != make_filepath(fn_buf, BUF_SIZE, dest, fi.filename))
      ret = SPI_ERROR_NOT_IMPLEMENTED;
    else
    {
      HANDLE  fh = CreateFile(fn_buf, GENERIC_WRITE, 0, NULL,
                              CREATE_NEW, FILE_ATTRIBUTE_NORMAL, NULL);
      if (INVALID_HANDLE_VALUE == fh)
        ret = SPI_ERROR_NOT_IMPLEMENTED;
      else
      {
        DWORD  write_bytes;
        WriteFile(fh, img, fi.filesize, &write_bytes, NULL);
        CloseHandle(fh);
      }
    }
    LocalFree(img);
  }

  return ret;
}

